#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <string.h>
#include <SDL/SDL.h>
#include <sys/time.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_image.h>
#include <unistd.h>

#define SMALL 0.0001

#ifndef FALSE
  #define FALSE 0
#endif
#ifndef TRUE
  #define TRUE 1
#endif


SDL_Surface *normalp;
SDL_Surface *texturep;

struct vector
{
  float x;
  float y;
  float z;
};

struct color
{
  float red;
  float green;
  float blue;
};

struct ray
{
  struct vector start;
  struct vector vector;
};

struct rect
{
  int x;
  int y;
  int width;
  int height;
};

struct light
{
  struct light *next;
  
  struct color color;
  struct vector center;
};

struct sphere
{
  struct sphere *next;
  
  struct color color;
  float reflect;
  
  struct vector center;
  float radius;
  struct rect bound;
};

struct camera
{
  struct vector position;
  struct vector look_at;
  int relative;
  float zoom;
};

struct camera_temp
{
  struct vector forward;
  struct vector up;
  struct vector left;
};

SDL_Surface *screen;

struct camera camera;
struct camera_temp camera_temp;
struct light *lights;
struct sphere *spheres;
struct render_thread_data *data;
int count = 0;
int threads = 1;
int quit = FALSE;
float ripple;

void inline normalize_vector(struct vector *vector)
{
  float tmp;
  tmp = 1/sqrt(vector->x*vector->x+vector->y*vector->y+vector->z*vector->z);
  vector->x *= tmp;
  vector->y *= tmp;
  vector->z *= tmp;
}

inline float get_mandelbrot(float x,float y)
{
  int i;
  float a = 0, b = 0, c, d;
  for (i = 0;i < 30;i++)
  {
    c = a*a-b*b+x;
    d = 2*a*b+y;
    a = c;
    b = d;
    if (a*a+b*b > 5)
      return i/30.0;
  }
  return 0.0;
}


void read_xml(char *file)
{
  xmlDoc *doc = NULL;
  struct sphere *sphere;
  struct light *light;
  xmlNode *node = NULL;
  xmlAttr *properties = NULL;
  float tmp;

  LIBXML_TEST_VERSION
  doc = xmlReadFile(file, NULL, 0);
  if (doc == NULL)
    exit (EXIT_FAILURE);
  lights = malloc(sizeof(struct light));
  light = lights;
  light->next = NULL;
  spheres = malloc(sizeof(struct sphere));
  sphere = spheres;
  sphere->next = NULL;
  for (node = xmlDocGetRootElement(doc)->children; node != NULL; node = node->next)
  {
    if (node->type == XML_ELEMENT_NODE)
    {
      if (xmlStrEqual(node->name, (xmlChar *) "light") == 1)
      {
        light->next = malloc(sizeof(struct light));
        light = light->next;
        light->next = NULL;
        for (properties = node->properties; properties; properties = properties->next)
        {
          tmp = atof((char *)properties->children->content);
          if (xmlStrEqual(properties->name, (xmlChar *) "x") == 1)
            light->center.x = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *) "y") == 1)
            light->center.y = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *) "z") == 1)
            light->center.z = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *) "red") == 1)
            light->color.red = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *) "green") == 1)
            light->color.green = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *) "blue") == 1)
            light->color.blue = tmp;
        }
      }
      else if (xmlStrEqual(node->name, (xmlChar *) "sphere") == 1)
      {
        sphere->next = malloc(sizeof(struct sphere));
        sphere = sphere->next;
        sphere->next = NULL;
        for (properties = node->properties; properties; properties = properties->next)
        {
          tmp = atof((char *)properties->children->content);
          if (xmlStrEqual(properties->name, (xmlChar *)"x") == 1)
            sphere->center.x = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"y") == 1)
            sphere->center.y = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"z") == 1)
            sphere->center.z = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"radius") == 1)
            sphere->radius = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"reflect") == 1)
            sphere->reflect = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"red") == 1)
            sphere->color.red = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"green") == 1)
            sphere->color.green = tmp;
          else if (xmlStrEqual(properties->name, (xmlChar *)"blue") == 1)
            sphere->color.blue = tmp;
        }
      }
    }
  }
  xmlFreeDoc(doc);
  xmlCleanupParser();
}

inline int intersect_sphere(float *a, float *b, struct sphere *sphere, struct ray *ray)
{
  float sx, sy, sz, tmpA, tmpB, tmpC;
  sx = sphere->center.x-ray->start.x;
  sy = sphere->center.y-ray->start.y;
  sz = sphere->center.z-ray->start.z;
  tmpA = ray->vector.x*sx+ray->vector.y*sy+ray->vector.z*sz;
  tmpB = ray->vector.x*ray->vector.x+ray->vector.y*ray->vector.y+ray->vector.z*ray->vector.z;
  tmpC = tmpA*tmpA-tmpB*(sx*sx+sy*sy+sz*sz-sphere->radius*sphere->radius);
  if (tmpC >= 0)
  {
    tmpC = sqrt (tmpC);
    *a = (tmpA+tmpC)/tmpB;
    *b = (tmpA-tmpC)/tmpB;
    return TRUE;
  }
  return FALSE;
}

void intersect_shapes(struct color *color, float *distance, struct ray *ray, int count, int scrx, int scry)
{
  struct sphere *sphere, *tmp_sphere;
  struct light *light;
  float a, b, tmp_dist;
  float sphere_distance = HUGE_VAL;
  tmp_sphere = NULL;
  struct color texcolor;
  //normalize_vector(&(ray->vector));
  for (sphere = spheres->next; sphere != NULL; sphere = sphere->next)
  {
    if (scrx != -1) {
      if (scrx < sphere->bound.x || scrx > sphere->bound.x+sphere->bound.width) continue;
      if (scry < sphere->bound.y || scry > sphere->bound.y+sphere->bound.height) continue;
    }
    if (intersect_sphere(&a, &b, sphere, ray))
    {
      if (a >= SMALL && a < sphere_distance)
      {
        tmp_sphere = sphere;
        sphere_distance = a;
      }
      if (b >= SMALL && b < sphere_distance)
      {
        tmp_sphere = sphere;
        sphere_distance = b;
      }
    }
  }
  struct color tmp_color;
  struct ray tmp_ray;
  if (color != NULL)
  {
    if (tmp_sphere == NULL)
    {
      color->red = 0;
      color->green = 0;
      color->blue = 0;
    }
    else
    {
      struct vector hit;
      hit.x = ray->start.x+ray->vector.x*sphere_distance;
      hit.y = ray->start.y+ray->vector.y*sphere_distance;
      hit.z = ray->start.z+ray->vector.z*sphere_distance;
      texcolor = tmp_sphere->color;
/*      if (tmp_sphere == spheres->next) {
        double ydiff = fabs(hit.y-(int)hit.y);
        double zdiff = fabs(hit.z-(int)hit.z);
        texcolor.red = 0;
        texcolor.green = 0;
        texcolor.blue = 0;
        if ((ydiff > .5 && zdiff > .5) || (ydiff < .5 && zdiff < .5))
        {
          texcolor.red=1;
          texcolor.green=1;
          texcolor.blue=1;
        }
      } */
      /* if (tmp_sphere == spheres->next && 0) {
        float t = get_mandelbrot(hit.y/3,hit.z/3);
        t *= t;
        texcolor.red=t;
        texcolor.green=t;
        texcolor.blue=t;
      } */
      int texx,texy;
      Uint8 r,g,b,q,z,y;
      if (tmp_sphere == spheres->next) {
        texx = hit.y*512;
        texy = hit.z*512;
        //texx += (15*sin((ripple*2*3.14159+texx*.1)));
        //texy += (10*sin(ripple*texy*.01));
        texx = texx % 512;
        texy = texy % 512;
        if (texx < 0) texx += 512;
        if (texy < 0) texy += 512;
        //if (texx < 0) texx = -texx;if (texy < 0) texy = -texy;
        Uint8 r,g,b;
        //printf("%i %i \n",texy, texx);
        SDL_GetRGB(*((Uint32*)texturep->pixels + texy*texturep->pitch/4+texx),texturep->format,&r,&g,&b);
        texcolor.red = r/256.0;
        texcolor.green = g/256.0;
        texcolor.blue = b/256.0;
      }
      color->red = .2*texcolor.red;
      color->green = .2*texcolor.green;
      color->blue = .2*texcolor.blue;
      struct vector normal;
      normal.x = hit.x-tmp_sphere->center.x;
      normal.y = hit.y-tmp_sphere->center.y;
      normal.z = hit.z-tmp_sphere->center.z;
      if (tmp_sphere == spheres->next) {
        Uint32 a = *((Uint32*)normalp->pixels + texy*normalp->pitch/4+texx);
        SDL_GetRGB(a,normalp->format,&r,&z,&y);
        if (texy == 511) SDL_GetRGB(*((Uint32*)normalp->pixels + 0*normalp->pitch/4+texx),normalp->format,&g,&z,&y);
        else SDL_GetRGB(*((Uint32*)normalp->pixels + (texy+1)*normalp->pitch/4+texx),normalp->format,&g,&z,&y);
        SDL_GetRGB(*((Uint32*)normalp->pixels + texy*normalp->pitch/4+texx),normalp->format,&b,&z,&y);
        if (texx == 511) SDL_GetRGB(*((Uint32*)normalp->pixels + texy*normalp->pitch/4+0),normalp->format,&q,&z,&y);
        else SDL_GetRGB(*((Uint32*)normalp->pixels + texy*normalp->pitch/4+texx+1),normalp->format,&q,&z,&y);
        normal.x = 128;
        normal.y = .1*(r-g);
        normal.z = .1*(b-q);
//        normal.z = .1*(r-128);
      }
      normalize_vector(&normal);
      for (light = lights->next; light != NULL; light = light->next)
      {
        struct vector lightv;
        lightv.x = light->center.x-hit.x;
        lightv.y = light->center.y-hit.y;
        lightv.z = light->center.z-hit.z;
        tmp_ray.start = hit;
        tmp_ray.vector = lightv;
        intersect_shapes(NULL,&tmp_dist,&tmp_ray,0,-1,-1);
        if (fabs(tmp_dist) < 1) continue;
        normalize_vector(&lightv);
        float amt = normal.x*lightv.x+normal.y*lightv.y+normal.z*lightv.z;
        if (amt < 0) amt=0;
        color->red += texcolor.red*amt*light->color.red;
        color->green += texcolor.green*amt*light->color.green;
        color->blue += texcolor.blue*amt*light->color.blue;
      }
      if (tmp_sphere->reflect > SMALL && count != 0) // r = o-2*o*n*dot(o*n)
      {
        struct vector inc_vect = ray->vector;
        normalize_vector(&inc_vect);
        float dot = inc_vect.x*normal.x+inc_vect.y*normal.y+inc_vect.z*normal.z;
        if (dot < -.1) {
          tmp_ray.start = hit;
          tmp_ray.vector.x = ray->vector.x-2*normal.x*dot;
          tmp_ray.vector.y = ray->vector.y-2*normal.y*dot;
          tmp_ray.vector.z = ray->vector.z-2*normal.z*dot;
          intersect_shapes (&tmp_color, NULL, &tmp_ray, count-1,-1,-1);
          color->red += tmp_color.red*tmp_sphere->reflect;
          color->green += tmp_color.green*tmp_sphere->reflect;
          color->blue += tmp_color.blue*tmp_sphere->reflect;
        }
      }
      if (color->red > 1) color->red = 1;
      if (color->green > 1) color->green = 1;
      if (color->blue > 1) color->blue = 1;
    }
  }
  if (distance != NULL)
    *distance = sphere_distance;
}

struct render_thread_data
{
  SDL_Thread *thread;
  int first;
  int interval;
  int status;
};

void render_frame(struct render_thread_data *data)
{
  int y;
  Uint32 *bufp;
  int x;
  struct ray ray;
  struct color color;
  ray.start = camera.position;
  for (y = data->first; y < screen->h; y = y + data->interval)
  {
    bufp = (Uint32 *) screen->pixels + y*screen->pitch/4;
    for (x = 0; x < screen->w; x++)
    {
      ray.vector.x = camera_temp.forward.x + camera_temp.left.x*(1-2*(float) x/screen->w) + camera_temp.up.x*(1-2*(float) y/screen->h);
      ray.vector.y = camera_temp.forward.y + camera_temp.left.y*(1-2*(float) x/screen->w) + camera_temp.up.y*(1-2*(float) y/screen->h);
      ray.vector.z = camera_temp.forward.z + camera_temp.left.z*(1-2*(float) x/screen->w) + camera_temp.up.z*(1-2*(float) y/screen->h);
      intersect_shapes (&color, NULL, &ray, 5, x, y);
      *(bufp+x) = SDL_MapRGB(screen->format, color.red*255, color.green*255, color.blue*255);
    }
  }
}

inline double get_time()
{
  struct timeval tv;
  gettimeofday (&tv, NULL);
  return tv.tv_sec+(double)tv.tv_usec/1000000.0;
}

inline void cross(struct vector *a,struct vector *b,struct vector *c)
{
  c->x = a->y*b->z-a->z*b->y;
  c->y = a->x*b->z-a->z*b->x;
  c->z = a->x*b->y-a->y*b->x;
}

struct vector up;

void loop_int()
{
  SDL_Event event;
  double time;
  int i;
  SDL_UpdateRect (screen, 0, 0, screen->w, screen->h);
  while (SDL_PollEvent(&event))
    if (event.type == SDL_QUIT)
    {
      quit = TRUE;
      printf("got quit\n");
    }
  time = get_time();
  camera.position.x = 0; //+10*(time/5 - floor(time/5));
  float angle = -2*3.1415926*(time/7 - floor(time/7)); // 3.14159/1.5+0*
  angle = 3.14159*.5;
  camera.position.y = 4*cos(angle);
  camera.position.z = 4*sin(angle);
  camera.look_at.x = 0;
  camera.look_at.y = 0;
  camera.look_at.z = 0;
  camera.relative = FALSE;
  camera.zoom = 1;
  if (camera.relative)
    camera_temp.forward = camera.look_at;
  else
  {
    camera_temp.forward.x = camera.look_at.x-camera.position.x;
    camera_temp.forward.y = camera.look_at.y-camera.position.y;
    camera_temp.forward.z = camera.look_at.z-camera.position.z;
  }
  normalize_vector(&camera_temp.forward);
  camera_temp.left.x = 0;
  camera_temp.left.y = cos(angle+3.14159/2);
  camera_temp.left.z = sin(angle+3.14159/2);
//  printf("%f %f %f --- ",camera_temp.left.x,camera_temp.left.y,camera_temp.left.z);
//  cross(&camera_temp.forward,&up,&camera_temp.left);
//  printf("%f %f %f\n",camera_temp.left.x,camera_temp.left.y,camera_temp.left.z);
//  normalize_vector(&camera_temp.left);
//  cross(&camera_temp.left,&camera_temp.forward,&camera_temp.up);
  camera_temp.up = up;
//  camera_temp.up.x = fabs(camera_temp.up.x);
//  normalize_vector(&camera_temp.up);
  camera_temp.forward.x *= camera.zoom;
  camera_temp.forward.y *= camera.zoom;
  camera_temp.forward.z *= camera.zoom;
  for (i=0;i<threads;i++)
    data[i].status = TRUE;
  count++;
  angle = -2*3.1415926*(time/7 - floor(time/7)); //1/1.5*3.1415926;
  lights->next->center.x=40;
  lights->next->center.y=100*cos(angle);
  lights->next->center.z=100*sin(angle);
  angle = 2*3.1415926*(time/3 - floor(time/3));
  spheres->next->next->center.x=fabs(1.5*sin(angle));
  angle = 2*3.1415926*(time/4 - floor(time/4));
  angle = -2*3.1415926*(time/7 - floor(time/7));
  spheres->next->next->center.y=2.5*cos(angle);
  spheres->next->next->center.z=2.5*sin(angle);
  ripple = (time/3 - floor(time/3));
  
  struct sphere *sphere;
for (sphere = spheres->next; sphere != NULL; sphere = sphere->next) {
/*  not given: y
        ray.vector.x = camera_temp.forward.x + camera_temp.left.x*(1-2*(float) x/screen->w) + camera_temp.up.x*(1-2*(float) y/screen->h);
      ray.vector.y = camera_temp.forward.y + camera_temp.left.y*(1-2*(float) x/screen->w) + camera_temp.up.y*(1-2*(float) y/screen->h);
      ray.vector.z = camera_temp.forward.z + camera_temp.left.z*(1-2*(float) x/screen->w) + camera_temp.up.z*(1-2*(float) y/screen->h);
  
    float sx, sy, sz, tmpA, tmpB, tmpC;
  sx = sphere->center.x-ray->start.x;
  sy = sphere->center.y-ray->start.y;
  sz = sphere->center.z-ray->start.z;
  tmpA = ray->vector.x*sx+ray->vector.y*sy+ray->vector.z*sz;
  tmpB = ray->vector.x*ray->vector.x+ray->vector.y*ray->vector.y+ray->vector.z*ray->vector.z;
  tmpC = tmpA*tmpA-tmpB*(sx*sx+sy*sy+sz*sz-sphere->radius*sphere->radius);
  tmpC = 0; */
    sphere->bound.x = 0;
    sphere->bound.y = 0;
    sphere->bound.width = 512;
    sphere->bound.height = 512;
    if (sphere != spheres->next && 0) {
        sphere->bound.x = 64;
    sphere->bound.y = 64;
    sphere->bound.width = 256;
    sphere->bound.height = 256;
    }
  }
}

void render_thread(struct render_thread_data *mydata)
{
  int i;
  int last;
  while (TRUE)
  {
    while (!mydata->status)
      SDL_Delay(1);
    render_frame(mydata);
    last = TRUE;
    for (i = 0;i < threads;i++)
      if (data[i].status && data[i].thread != mydata->thread)
        last = FALSE;
    if (last)
      loop_int();
    else
      mydata->status = FALSE;
    if (quit)
      break;
  }
}

int main(int argc, char **argv)
{
  int xsize = 512;
  int ysize = 512;
  int i;
  double start;
  
  up.x = 1.0;
  up.y = 0.0;
  up.z = 0.0;
  
  if (argc != 2 && argc != 5)
  {
    printf("Usage: %s <scenefile> [threads xsize ysize]\n", argv[0]);
    return EXIT_FAILURE;
  }
  if (argc == 5)
  {
    threads = atoi(argv[2]);
    xsize = atoi(argv[3]);
    ysize = atoi(argv[4]);
  }
  read_xml(argv[1]);
  SDL_Init (SDL_INIT_VIDEO);
  screen = SDL_SetVideoMode (xsize, ysize, 32, SDL_SWSURFACE);
  if (screen == NULL || screen->offset || ((screen->flags & (SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_RLEACCEL)) != 0))
  {
    printf ("Some kind of SDL failure...\n");
    return EXIT_FAILURE;
  }
  data = malloc(threads*sizeof(struct render_thread_data));
  
  for (i = 0; i < threads; i++)
  {
    data[i].first = i;
    data[i].interval = threads;
    data[i].status = FALSE;
    if (i != 0)
      data[i].thread = SDL_CreateThread((void *)render_thread, data+i);
  }
  normalp = SDL_DisplayFormat(IMG_Load("normal.jpg"));
  texturep = SDL_DisplayFormat(IMG_Load("texture.jpg"));
  start = get_time();
  loop_int();

  render_thread(data);
  
  printf ("%f\n",count/(get_time()-start));
  for (i = 1; i < threads; i++)
    SDL_WaitThread(data[i].thread, NULL);
  SDL_Quit();
  return EXIT_SUCCESS;
}
